import logging
import os
import shlex
import subprocess
import time

import dask.dataframe as dd
import numpy as np
import pytest

from spatialpandas import GeoDataFrame, GeoSeries
from spatialpandas.geometry import PointArray
from spatialpandas.io import read_parquet, read_parquet_dask, to_parquet, to_parquet_dask


pytest.importorskip("moto")
geopandas = pytest.importorskip("geopandas")
s3fs = pytest.importorskip("s3fs")
requests = pytest.importorskip("requests")

logging.getLogger("botocore").setLevel(logging.INFO)


@pytest.fixture(scope="module", autouse=True)
def s3_fixture():
    """Writable local S3 system.

    Taken from `universal_pathlib/upath/tests` and `s3fs/tests/test_s3fs.py`.
    """
    if "BOTO_CONFIG" not in os.environ:  # pragma: no cover
        os.environ["BOTO_CONFIG"] = "/dev/null"
    if "AWS_ACCESS_KEY_ID" not in os.environ:  # pragma: no cover
        os.environ["AWS_ACCESS_KEY_ID"] = "testing"
    if "AWS_SECRET_ACCESS_KEY" not in os.environ:  # pragma: no cover
        os.environ["AWS_SECRET_ACCESS_KEY"] = "testing"
    if "AWS_SECURITY_TOKEN" not in os.environ:  # pragma: no cover
        os.environ["AWS_SECURITY_TOKEN"] = "testing"
    if "AWS_SESSION_TOKEN" not in os.environ:  # pragma: no cover
        os.environ["AWS_SESSION_TOKEN"] = "testing"
    if "AWS_DEFAULT_REGION" not in os.environ:  # pragma: no cover
        os.environ["AWS_DEFAULT_REGION"] = "us-east-1"

    port = 5555
    bucket_name = "test_bucket"
    endpoint_url = f"http://127.0.0.1:{port}/"
    proc = subprocess.Popen(
        shlex.split(f"moto_server s3 -p {port}"),
        stderr=subprocess.DEVNULL,
        stdout=subprocess.DEVNULL,
    )
    try:
        timeout = 5
        while timeout > 0:
            try:
                r = requests.get(endpoint_url, timeout=10)
                if r.ok:
                    break
            except Exception:  # pragma: no cover
                pass
            timeout -= 0.1  # pragma: no cover
            time.sleep(0.1)  # pragma: no cover
        anon = False
        s3so = {
            "anon": anon,
            "endpoint_url": endpoint_url,
        }
        fs = s3fs.S3FileSystem(**s3so)
        fs.mkdir(bucket_name)
        assert fs.exists(bucket_name)
        yield f"s3://{bucket_name}", s3so
    finally:
        proc.terminate()
        proc.wait()


@pytest.fixture(scope="module")
def sdf():
    src_array = np.array([[0, 1], [2, 3], [4, 5], [6, 7]], dtype=np.float32)
    points = PointArray(src_array)
    return GeoDataFrame({"point": GeoSeries(points)})


@pytest.fixture(scope="module")
def s3_parquet_dask(s3_fixture, sdf):
    path, s3so = s3_fixture
    path = f"{path}/test_dask"
    ddf = dd.from_pandas(sdf, npartitions=2)
    to_parquet_dask(ddf, path, storage_options=s3so)
    fs = s3fs.S3FileSystem(**s3so)
    assert fs.exists(path) and fs.isdir(path)
    yield path, s3so, sdf


@pytest.fixture(scope="module")
def s3_parquet_pandas(s3_fixture, sdf):
    path, s3so = s3_fixture
    path = f"{path}/test_pandas.parquet"
    to_parquet(sdf, path, storage_options=s3so)
    fs = s3fs.S3FileSystem(**s3so)
    assert fs.exists(path) and fs.isfile(path)
    yield path, s3so, sdf


class TestS3ParquetDask:
    @staticmethod
    def test_read_parquet_dask_remote_glob_parquet(s3_parquet_dask):
        path, s3so, sdf = s3_parquet_dask
        result = read_parquet_dask(f"{path}/*.parquet", storage_options=s3so).compute()
        assert result.equals(sdf)

    @staticmethod
    def test_read_parquet_dask_remote_glob_all(s3_parquet_dask):
        path, s3so, sdf = s3_parquet_dask
        result = read_parquet_dask(f"{path}/*", storage_options=s3so).compute()
        assert result.equals(sdf)

    @staticmethod
    def test_read_parquet_dask_remote_dir(s3_parquet_dask):
        path, s3so, sdf = s3_parquet_dask
        result = read_parquet_dask(path, storage_options=s3so).compute()
        assert result.equals(sdf)

    @staticmethod
    def test_read_parquet_dask_remote_dir_slash(s3_parquet_dask):
        path, s3so, sdf = s3_parquet_dask
        result = read_parquet_dask(f"{path}/", storage_options=s3so).compute()
        assert result.equals(sdf)


def test_read_parquet_remote(s3_parquet_pandas):
    path, s3so, sdf = s3_parquet_pandas
    result = read_parquet(path, storage_options=s3so)
    assert result.equals(sdf)



import numpy as np

from spatialpandas.geometry import PointArray


def test_construct_pointarray_interleaved():
    src_array = np.array([0, 1, 2, 3, 4, 5, 6, 7], dtype=np.float32)
    points = PointArray(src_array)

    np.testing.assert_array_equal(points.x, src_array[0::2])
    np.testing.assert_array_equal(points.y, src_array[1::2])
    np.testing.assert_array_equal(points.isna(), np.isnan(src_array[0::2]))
    np.testing.assert_array_equal(points.flat_values, src_array)


def test_construct_pointarray_2d():
    src_array = np.array([[0, 1], [2, 3], [4, 5], [6, 7]], dtype=np.float32)
    points = PointArray(src_array)

    np.testing.assert_array_equal(points.x, src_array[:, 0])
    np.testing.assert_array_equal(points.y, src_array[:, 1])
    np.testing.assert_array_equal(points.isna(), np.isnan(src_array[:, 0]))
    np.testing.assert_array_equal(points.flat_values, src_array.flatten())


def test_construct_pointarray_tuple():
    src_array = np.array([[0, 1], [2, 3], [4, 5], [6, 7]], dtype=np.float32)
    points = PointArray((src_array[:, 0], src_array[:, 1]))

    np.testing.assert_array_equal(points.x, src_array[:, 0])
    np.testing.assert_array_equal(points.y, src_array[:, 1])
    np.testing.assert_array_equal(points.isna(), np.isnan(src_array[:, 0]))
    np.testing.assert_array_equal(points.flat_values, src_array.flatten())


def test_construct_pointarray_2d_with_None():
    src_array = np.array([None, [2, 3], [4, 5], None], dtype=object)
    expected_xs = np.array([np.nan, 2, 4, np.nan], dtype=np.float64)
    expected_ys = np.array([np.nan, 3, 5, np.nan], dtype=np.float64)

    points = PointArray(src_array)

    np.testing.assert_array_equal(points.x, expected_xs)
    np.testing.assert_array_equal(points.y, expected_ys)
    np.testing.assert_array_equal(points.isna(), np.isnan(expected_xs))
    np.testing.assert_array_equal(
        points.flat_values[2:6], np.array([2, 3, 4, 5], dtype=np.int64)
    )



from functools import total_ordering

import numpy as np
import pyarrow as pa
from numba import jit, prange

from ..geometry.base import Geometry, GeometryArray
from ._algorithms.bounds import (bounds_interleaved, total_bounds_interleaved,
                                 total_bounds_interleaved_1d)


def _validate_nested_arrow_type(nesting_levels, pyarrow_type):
    if pyarrow_type == pa.null():
        return pa.null()

    pyarrow_element_type = pyarrow_type
    for i in range(nesting_levels):
        if not isinstance(pyarrow_element_type, pa.ListType):
            raise ValueError(
                "Expected input data to have {} nested layer(s)".format(
                    nesting_levels)
            )
        pyarrow_element_type = pyarrow_element_type.value_type
    pyarrow_element_type = pyarrow_element_type
    numpy_element_dtype = pyarrow_element_type.to_pandas_dtype()
    if (numpy_element_dtype() is None
            or numpy_element_dtype().dtype.kind not in ('i', 'u', 'f')):
        raise ValueError(
            "Invalid nested element type {}, expected numeric type".format(
                pyarrow_element_type
            ))
    return pyarrow_element_type


class _ListArrayBufferMixin:
    """
    Mixin of buffer utilities for classes that store a pyarrow ListArray as their
    listarray property. The numpy data type of the inner ListArray elements must be
    stored as the numpy_dtype property
    """
    @property
    def buffer_values(self):
        value_buffer = self.listarray.buffers()[-1]
        if value_buffer is None:
            return np.array([], dtype=self.numpy_dtype)
        else:
            return np.asarray(value_buffer).view(self.numpy_dtype)

    @property
    def buffer_offsets(self):
        """
        Tuple of offsets arrays, one for each tested level.
        """
        buffers = self.listarray.buffers()
        if len(buffers) < 2:
            return (np.array([0]),)
        elif len(buffers) < 3:
            # offset values that include everything
            return (np.array([0, len(self.listarray)]),)

        # Slice first offsets array to match any current extension array slice
        # All other buffers remain unchanged
        start = self.listarray.offset
        stop = start + len(self.listarray) + 1
        offsets1 = np.asarray(buffers[1]).view(np.uint32)[start:stop]

        remaining_offsets = tuple(
            np.asarray(buffers[i]).view(np.uint32)
            for i in range(3, len(buffers) - 1, 2)
        )

        return (offsets1,) + remaining_offsets

    @property
    def flat_values(self):
        """
        Flat array of the valid values. This differs from buffer_values if the pyarrow
        ListArray backing this object is a slice. buffer_values will contain all
        values from the original (pre-sliced) object whereas flat_values will contain
        only the sliced values.
        """
        # Compute valid start/stop index into buffer values array.
        buffer_offsets = self.buffer_offsets
        start = buffer_offsets[0][0]
        stop = buffer_offsets[0][-1]
        for offsets in buffer_offsets[1:]:
            start = offsets[start]
            stop = offsets[stop]

        return self.buffer_values[start:stop]

    @property
    def buffer_outer_offsets(self):
        """
        Array of the offsets into buffer_values that separate the outermost nested
        structure of geometry object(s), regardless of the number of nesting levels.
        """
        buffer_offsets = self.buffer_offsets
        flat_offsets = buffer_offsets[0]
        for offsets in buffer_offsets[1:]:
            flat_offsets = offsets[flat_offsets]

        return flat_offsets

    @property
    def buffer_inner_offsets(self):
        """
        Array of the offsets into buffer_values that separate the innermost nested
        structure of geometry object(s), regardless of the number of nesting levels.
        """
        buffer_offsets = self.buffer_offsets
        start = buffer_offsets[0][0]
        stop = buffer_offsets[0][-1]
        for offsets in buffer_offsets[1:-1]:
            start = offsets[start]
            stop = offsets[stop]
        return buffer_offsets[-1][start:stop + 1]


@total_ordering
class GeometryList(Geometry, _ListArrayBufferMixin):
    """
    Base class for elements of GeometryListArray subclasses
    """
    _nesting_levels = 0

    @staticmethod
    def _pa_element_value_type(data):
        """
        Get value type of pyarrow ListArray element for different versions of pyarrow
        """
        try:
            # Try pyarrow 1.0 API
            return data.values.type
        except AttributeError:
            # Try pre 1.0 API
            return data.value_type

    @staticmethod
    def _pa_element_values(data):
        """
        Get values of nested pyarrow ListArray element for different versions of pyarrow
        """
        try:
            # Try pyarrow 1.0 API
            return data.values
        except AttributeError:
            # Try pre 1.0 API
            return pa.array(data.as_py(), data.value_type)

    def __init__(self, data, dtype=None):
        super().__init__(data)
        if len(self.data) > 0:
            value_type = GeometryList._pa_element_value_type(self.data)
            _validate_nested_arrow_type(self._nesting_levels, value_type)

        # create listarray for _ListArrayBufferMixin
        self.listarray = GeometryList._pa_element_values(self.data)

    def __lt__(self, other):
        if type(other) is not type(self):
            return NotImplemented
        return _lexographic_lt(np.asarray(self.listarray), np.asarray(other.listarray))

    def __len__(self):
        return len(self.buffer_outer_offsets - 1)

    @classmethod
    def construct_array_type(cls):
        return GeometryListArray

    @property
    def numpy_dtype(self):
        if isinstance(self.listarray, pa.NullArray):
            return None
        else:
            typ = GeometryList._pa_element_value_type(self.data)
            for _ in range(self._nesting_levels):
                typ = GeometryList._pa_element_value_type(typ)
            return np.dtype(typ.to_pandas_dtype())


class GeometryListArray(GeometryArray, _ListArrayBufferMixin):
    """
    Base class for geometry arrays that are backed by a pyarrow ListArray.
    """
    _element_type = GeometryList
    _nesting_levels = 1

    @classmethod
    def _arrow_type_from_numpy_element_dtype(cls, dtype):
        # Scalar element dtype
        arrow_dtype = pa.from_numpy_dtype(dtype)

        # Wrap dtype with appropriate number of nesting levels
        for i in range(cls._nesting_levels):
            arrow_dtype = pa.list_(arrow_dtype)

        return arrow_dtype

    def _numpy_element_dtype_from_arrow_type(self, pyarrow_type):
        if pyarrow_type == pa.null():
            return pa.null()

        pyarrow_element_type = pyarrow_type
        for i in range(self._nesting_levels):
            pyarrow_element_type = pyarrow_element_type.value_type

        return pyarrow_element_type.to_pandas_dtype()

    # Constructor
    def __init__(self, array, dtype=None):
        super().__init__(array, dtype)

        # Set listarray property for _ListArrayBufferMixin
        self.listarray = self.data

        # Check that inferred type has the right number of nested levels
        _validate_nested_arrow_type(
            self._nesting_levels, self.data.type
        )

        # Validate input data is compatible
        offsets = self.buffer_offsets

        # Validate even number of inner elements per polygon
        if any((offsets[-1] % 2) > 0):
            raise ValueError("""
GeometryList objects are represented by interleaved x and y coordinates, so they must have
an even number of elements. Received specification with an odd number of elements.""")

    # Base geometry methods
    @property
    def total_bounds(self):
        return total_bounds_interleaved(self.flat_values)

    @property
    def total_bounds_x(self):
        return total_bounds_interleaved_1d(self.flat_values, 0)

    @property
    def total_bounds_y(self):
        return total_bounds_interleaved_1d(self.flat_values, 1)

    @property
    def bounds(self):
        return bounds_interleaved(self.buffer_values, self.buffer_outer_offsets)


@jit(nopython=True, nogil=True)
def _lexographic_lt0(a1, a2):
    """
    Compare two 1D numpy arrays lexographically
    Parameters
    ----------
    a1: ndarray
        1D numpy array
    a2: ndarray
        1D numpy array

    Returns
    -------
    comparison:
        True if a1 < a2, False otherwise
    """
    for e1, e2 in zip(a1, a2):
        if e1 < e2:
            return True
        elif e1 > e2:
            return False
    return len(a1) < len(a2)


def _lexographic_lt(a1, a2):
    if a1.dtype != np.dtype(object) and a1.dtype != np.dtype(object):
        # a1 and a2 primitive
        return _lexographic_lt0(a1, a2)
    elif a1.dtype == np.dtype(object) and a1.dtype == np.dtype(object):
        # a1 and a2 object, process recursively
        for e1, e2 in zip(a1, a2):
            if _lexographic_lt(e1, e2):
                return True
            elif _lexographic_lt(e2, e1):
                return False
        return len(a1) < len(a2)
    elif a1.dtype != np.dtype(object):
        # a2 is object array, a1 primitive
        return True
    else:
        # a1 is object array, a2 primitive
        return False


@jit(nogil=True, nopython=True, parallel=True)
def _geometry_map_nested1(
        fn, result, values, value_offsets, missing
):
    assert len(value_offsets) == 1
    value_offsets0 = value_offsets[0]
    n = len(value_offsets0) - 1
    for i in prange(n):
        if not missing[i]:
            result[i] = fn(values, value_offsets0[i:i + 2])


@jit(nogil=True, nopython=True, parallel=True)
def _geometry_map_nested2(
        fn, result, values, value_offsets, missing
):
    assert len(value_offsets) == 2
    value_offsets0 = value_offsets[0]
    value_offsets1 = value_offsets[1]
    n = len(value_offsets0) - 1
    for i in prange(n):
        if not missing[i]:
            start = value_offsets0[i]
            stop = value_offsets0[i + 1]
            result[i] = fn(values, value_offsets1[start:stop + 1])


@jit(nogil=True, nopython=True, parallel=True)
def _geometry_map_nested3(
        fn, result, values, value_offsets, missing
):
    assert len(value_offsets) == 3
    value_offsets0 = value_offsets[0]
    value_offsets1 = value_offsets[1]
    value_offsets2 = value_offsets[2]
    n = len(value_offsets0) - 1
    for i in prange(n):
        if not missing[i]:
            start = value_offsets1[value_offsets0[i]]
            stop = value_offsets1[value_offsets0[i + 1]]
            result[i] = fn(values, value_offsets2[start:stop + 1])



from hypothesis import given
from shapely import geometry as sg

from ..strategies import coord, hyp_settings
from spatialpandas.geometry._algorithms.orientation import triangle_orientation


@given(coord, coord, coord, coord, coord, coord)
@hyp_settings
def test_triangle_orientation(ax, ay, bx, by, cx, cy):
    result = triangle_orientation(ax, ay, bx, by, cx, cy)

    # Use shapely polygon to compute expected orientation
    sg_poly = sg.Polygon([(ax, ay), (bx, by), (cx, cy), (ax, ay)])

    if sg_poly.area == 0:
        expected = 0
    else:
        expected = 1 if sg_poly.exterior.is_ccw else -1

    assert result == expected



import numpy as np
import pytest
from hypothesis import given

from ..strategies import hyp_settings, st_multipolygon_array, st_polygon_array
from spatialpandas.geometry import MultiPolygonArray, PolygonArray


@pytest.mark.slow
@given(st_polygon_array())
@hyp_settings
def test_polygon_area(gp_polygon):
    polygons = PolygonArray.from_geopandas(gp_polygon)
    expected_area = gp_polygon.area
    area = polygons.area
    np.testing.assert_allclose(area, expected_area)


@pytest.mark.slow
@given(st_multipolygon_array())
@hyp_settings
def test_multipolygon_area(gp_multipolygon):
    multipolygons = MultiPolygonArray.from_geopandas(gp_multipolygon)
    expected_area = gp_multipolygon.area
    area = multipolygons.area
    np.testing.assert_allclose(area, expected_area)



import dask.dataframe as dd
import pandas as pd

import spatialpandas as sp


def test_dask_registration():
    ddf = dd.from_pandas(sp.GeoDataFrame({
        'geom': pd.array(
            [[0, 0], [0, 1, 1, 1], [0, 2, 1, 2, 2, 2]], dtype='MultiPoint[float64]'),
        'v': [1, 2, 3]
    }), npartitions=3)
    assert isinstance(ddf, sp.dask.DaskGeoDataFrame)
